# -*- coding: utf-8 -*-
##############################################################################
#
#    NUMA
#    Copyright (C) 2011 NUMA Extreme Systems (<http:www.numaes.com>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta
import time

from osv import fields, osv
from tools.translate import _
import decimal_precision as dp
import netsvc
import traceback
import re

import pdb


class configurator(osv.osv_memory):
    _name = "product.conf_wizard"

    def _get_product_domain (self, cr, uid, ids, fields=None, args=None, context=None):
        context = context or {}

        res = {}
        for c in self.browse (cr, uid, ids, context=context): 
            if c.dimension_id and c.dimension_id.atom_type == 'product':
                e_context = dict (context, {'product': c})
                domain = eval(c.dimension_id.product_domain, e_context)
                res[c.id] = domain or []
            else:
                res[c.id] = []
        return res

    _columns = {
        'product_id': fields.many2one ('product.product', 'Product to configure'),
        'dimension_domain': fields.related ('product_id', 'conf_all_dimensions_ids', type="one2many", obj="product.conf_dimension", string="Dimension domain"),
        'values_ids': fields.related ('product_id', 'conf_all_values_ids', type="one2many", obj="product.conf_value", string="All current values"),

        'enabled': fields.boolean ('Required?'),

        'dimension_id': fields.many2one ('product.conf_dimension', 'Current dimension to configure'),
        'atom_type': fields.related ('dimension_id', 'atom_type', type="char", size=32, string='Atom type', readonly=True), 
        'dimension_image': fields.related ('dimension_id', 'picture', type="binary", filters="*.png,*.jpg,*.bmp", string='Dimension image'), 
        'selection_domain': fields.related ('dimension_id', 'selection_list_ids', type="one2many", obj="product.conf_selection_value", string="Selection domain"),
        'product_domain': fields.function (_get_product_domain, method=True, type="one2many", obj="product.product", string="Product domain"),
        'value_id': fields.many2one ('product.conf_value', 'Current value to set'),

        'name': fields.related ('dimension_id', 'name', strint='Name', type="char"),
        'description': fields.related ('dimension_id', 'description', string='Prompt description', type="text"),

        'number': fields.float ('As number', digits=(16,2)),
        'text': fields.text ('As text or selection'),
        'currency_id': fields.many2one ('res.currency', 'As currency'),
        'date': fields.date ('As date'),
        'timestamp': fields.datetime ('As Point in Time'),
        'time': fields.time ('As elapsed time'),
        'uom_id': fields.many2one ('product.uom', 'As UoM'),
        'selected_product_id': fields.many2one ('product.product', 'As product'),
        'current_selection_id': fields.many2one ('product.conf_selection_value', 'Current selection'),

    }

    def onchange_dimension_id (self, cr, uid, ids, dimension_id, product_id, context=None):
        context = context or self.get_user_context(cr, uid)
        dimension_obj = self.pool.get('product.conf_dimension')
        product_obj = self.pool.get('product.product')
        value_obj = self.pool.get('product.conf_value')

        if dimension_id:
            dimension = dimension_obj.browse (cr, uid, dimension_id, context=context)
            product = product_obj.browse (cr, uid, product_id, context=context)

            value_ids = product.get_all_conf_values_ids ()[product.id]
            value_ids = value_obj.search (cr, uid, [('dimension_id','=',dimension_id),('id','in', value_ids)], context=context)
            vals = {}
            wvals = {}
            if not value_ids:
                value_id = False
            else:
                value_id = value_ids[0]
                value = value_obj.browse (cr, uid, value_id, context=context)
                vals = self.fill_values (cr, uid, value.get_value(context=context), dimension)
                wvals = vals
            wvals['dimension_id'] = dimension_id
            wvals['value_id'] = value_id
            self.write (cr, uid, ids, wvals, context=context)

            res = {
                'atom_type': dimension.atom_type,
                'dimension_image': dimension.picture,
                'dimension_image': '',
                'selection_domain': [sd.id for sd in dimension.selection_list_ids],
                'name': dimension.name,
                'description': dimension.description,
                'values_ids': [v.id for v in product.conf_all_values_ids],
                'product_domain': self._get_product_domain(cr, uid, ids, context=context)[ids[0]],
            }

            res.update(vals)

            return {'value': res}
        else:
            return False


    def fill_values (self, cr, uid, current_value, selected_dimension):
        selection_value_obj = self.pool.get('product.conf_selection_value')

        res = {
            'number': 0.0,
            'text': '',
            'currency_id': selected_dimension and selected_dimension.default_currency_id.id,
            'date': False,
            'timestamp': False,
            'time': False,
            'uom_id': False,
            'selected_product_id': False,
            'current_selection_id': False,
        }

        if current_value and selected_dimension:
            if selected_dimension.atom_type == 'number':
                res ['number'] = current_value
            elif selected_dimension.atom_type == 'text':
                res ['text'] = str(current_value)
            elif selected_dimension.atom_type == 'currency':
                res ['currency_id'] = isinstance(current_value, int) and current_value or current_value.id
            elif selected_dimension.atom_type == 'date':
                res ['date'] = current_value
            elif selected_dimension.atom_type == 'timestamp':
                res ['timestamp'] = current_value
            elif selected_dimension.atom_type == 'time':
                res ['time'] = current_value
            elif selected_dimension.atom_type == 'measure':
                res ['number'] = v['number']
                res ['uom_id'] = isinstance(current_value['uom_id'], int) and current_value['uom_id'] or current_value['uom_id'].id
            elif selected_dimension.atom_type == 'product':
                res ['selected_product_id'] = isinstance(current_value, int) and current_value or current_value.id
            elif selected_dimension.atom_type == 'selection':
                res ['current_selection_id'] = isinstance(current_value, int) and current_value or current_value.id

        return res


    def create (self, cr, uid, vals, context=None):
        context = context or {}
        product_obj = self.pool.get ('product.product')
        value_obj = self.pool.get ('product.conf_value')

        new_id = super (configurator, self).create (cr, uid, vals, context)

        new_vals = vals.copy()
        if vals.get ('product_id', False):
            #Initializing the configurator to a new wizard
            product = product_obj.browse (cr, uid,  vals['product_id'], context=context)
            dimensions, dimensions_mask = product.conf_mask()
            if dimensions:
                selected_dimension = dimensions[0]

                value_ids = product.get_all_conf_values_ids ()
                value_ids = value_obj.search (cr, uid, [('dimension_id','=',selected_dimension.id),('id','in', value_ids)], context=context)
                if not value_ids:
                    value_id = False
                else:
                    value_id = value_ids[0]
                    value = value_obj.browse (cr, uid, value_id, context=context)
                    new_vals = self.fill_values (cr, uid, value.get_value(context=context), selected_dimension)
                new_vals['dimension_id'] = selected_dimension and selected_dimension.id or False
                new_vals['value_id'] = value_id
                new_vals['enabled'] = selected_dimension and dimensions_mask[selected_dimension.id] or True,

        if new_vals:
            self.write (cr, uid, [new_id], new_vals, context=context)

        return new_id


    def refresh_value (self, cr, uid, ids, context=None):
        context = context or {}
        value_obj = self.pool.get ('product.conf_value')

        assert ids and len(ids)==1

        configurator = self.browse (cr, uid, ids[0], context=context)

        vals = {
            'number': configurator.number,
            'text': configurator.text,
            'currency_id': configurator.currency_id.id,
            'date': configurator.date,
            'timestamp': configurator.timestamp,
            'time': configurator.time,
            'uom_id': configurator.uom_id.id,
            'selected_product_id': configurator.selected_product_id.id,
            'current_selection_id': configurator.current_selection_id.id,
        }

        if configurator.value_id and configurator.value_id.product_id.id == configurator.product_id.id:
            value_obj.write (cr, uid, [configurator.value_id.id], vals, context=context)
        elif configurator.dimension_id:
            if configurator.dimension_id:
                vals['product_id'] = configurator.product_id.id
                vals['dimension_id'] = configurator.dimension_id.id
                value_id = value_obj.create (cr, uid, vals, context=context)
                self.write (cr, uid, ids, {'value_id': value_id}, context=context)

    def get_user_context (self, cr, uid):
        # Patch for GTK client fault: no context on onchange
        user_obj = self.pool.get ('res.users')

        user = user_obj.browse (cr, uid, uid)
        return {'lang': user.context_lang, 'tz': user.context_tz}

    def onchange_number (self, cr, uid, ids, number, context=None):
        context = context or self.get_user_context(cr, uid)

        assert ids and len(ids)==1

        configurator = self.browse (cr, uid, ids[0], context=context)

        assert configurator.atom_type in ['number', 'measure']

        if configurator.dimension_id.min_number > number:
            raise osv.except_osv (_('Value error'),
                   _('Dimension [%(name)s] does not accept a number lower than [%(minimun)d]. Please check!') % {
                            'name': configurator.dimension_id.name,
                            'minimun': configurator.dimension_id.min_number})

        if configurator.dimension_id.max_number >= configurator.dimension_id.min_number and number > configurator.dimension_id.max_number:
            raise osv.except_osv (_('Value error'),
                   _('Dimension [%(name)s] does not accept a number greater than [%(maximun)d]. Please check!') % {
                            'name': configurator.dimension_id.name,
                            'maximun': configurator.dimension_id.max_number})

        configurator.product_id.conf_validate()

        self.write (cr, uid, ids, {'number': number}, context=context)
        configurator.refresh_value()

        return {'value': {
                    'values_ids': [v.id for v in configurator.product_id.get_all_values()],
                }}

    def onchange_text (self, cr, uid, ids, text, context=None):
        context = context or self.get_user_context(cr, uid)

        assert ids and len(ids)==1

        configurator = self.browse (cr, uid, ids[0], context=context)

        assert configurator.atom_type in ['text']

        if configurator.dimension_id.text_mask:
            tm = re.compile (configurator.dimension_id.text_mask)
        
            if not tm.match(text):
                raise osv.except_osv (_('Value error'),
                       _('Dimension [%(name)s] does not accept this text (It fails the check against the allowed expression). Please check!') % {
                                'name': configurator.dimension_id.name,
                                'text_mask': configurator.dimension_id.text_mask})

        configurator.product_id.conf_validate()

        self.write (cr, uid, ids, {'text': text}, context=context)
        configurator.refresh_value()

        return {'value': {
                    'values_ids': [v.id for v in configurator.product_id.get_all_values()],
                }}

    def onchange_currency_id (self, cr, uid, ids, currency_id, context=None):
        context = context or self.get_user_context(cr, uid)

        assert ids and len(ids)==1

        configurator = self.browse (cr, uid, ids[0], context=context)

        assert configurator.atom_type == 'currency'

        configurator.product_id.conf_validate()

        self.write (cr, uid, ids, {'currency_id': curency_id}, context=context)
        configurator.refresh_value()

        return {'value': {
                    'values_ids': [v.id for v in configurator.product_id.get_all_values()],
                }}

    def onchange_date (self, cr, uid, ids, date, context=None):
        context = context or self.get_user_context(cr, uid)

        assert ids and len(ids)==1

        configurator = self.browse (cr, uid, ids[0], context=context)

        assert configurator.atom_type == 'date'

        if configurator.dimension_id.earliest_date and date < configurator.dimension_id.earliest_date:
            raise osv.except_osv (_('Value error'),
                   _('Dimension [%(name)s] does not accept a date earlier than [%(date)s]. Please check!') % {
                            'name': configurator.dimension_id.name,
                            'date': configurator.dimension_id.earliest_date})

        if configurator.dimension_id.latest_date and date > configurator.dimension_id.latest_date:
            raise osv.except_osv (_('Value error'),
                   _('Dimension [%(name)s] does not accept a date later than [%(date)s]. Please check!') % {
                            'name': configurator.dimension_id.name,
                            'date': configurator.dimension_id.earliest_date})

        configurator.product_id.conf_validate()

        self.write (cr, uid, ids, {'date': date}, context=context)
        configurator.refresh_value()

        return {'value': {
                    'values_ids': [v.id for v in configurator.product_id.get_all_values()],
                }}

    def onchange_timestamp (self, cr, uid, ids, timestamp, context=None):
        context = context or self.get_user_context(cr, uid)

        assert ids and len(ids)==1

        configurator = self.browse (cr, uid, ids[0], context=context)

        assert configurator.atom_type == 'timestamp'

        if configurator.dimension_id.earliest_point_in_time and timestamp < configurator.dimension_id.earliest_point_in_time:
            raise osv.except_osv (_('Value error'),
                   _('Dimension [%(name)s] does not accept a point in time earlier than [%(date)s]. Please check!') % {
                            'name': configurator.dimension_id.name,
                            'date': configurator.dimension_id.earliest_point_in_time})

        if configurator.dimension_id.latest_date and timestamp > configurator.dimension_id.latest_point_in_time:
            raise osv.except_osv (_('Value error'),
                   _('Dimension [%(name)s] does not accept a point in time later than [%(date)s]. Please check!') % {
                            'name': configurator.dimension_id.name,
                            'date': configurator.dimension_id.earliest_point_in_time})

        configurator.product_id.conf_validate()

        self.write (cr, uid, ids, {'timestamp': timestamp}, context=context)
        configurator.refresh_value()

        return {'value': {
                    'values_ids': [v.id for v in configurator.product_id.get_all_values()],
                }}

    def onchange_time (self, cr, uid, ids, time, context=None):
        context = context or self.get_user_context(cr, uid)

        assert ids and len(ids)==1

        configurator = self.browse (cr, uid, ids[0], context=context)

        assert configurator.atom_type == 'time'

        if configurator.dimension_id.max_time_lenght and time > configurator.dimension_id.max_time_lenght:
            raise osv.except_osv (_('Value error'),
                   _('Dimension [%(name)s] does not accept a time greater than [%(time)s]. Please check!') % {
                            'name': configurator.dimension_id.name,
                            'time': configurator.dimension_id.max_time_lenght})

        configurator.product_id.conf_validate()

        self.write (cr, uid, ids, {'time': time}, context=context)
        configurator.refresh_value()

        return {'value': {
                    'values_ids': [v.id for v in configurator.product_id.get_all_values()],
                }}

    def onchange_uom_id (self, cr, uid, ids, uom_id, context=None):
        context = context or self.get_user_context(cr, uid)
        uom_obj = self.pool.get("product.uom")
        uom_categ_obj = self.pool.get ("product.uom.categ")

        assert ids and len(ids)==1

        configurator = self.browse (cr, uid, ids[0], context=context)

        assert configurator.atom_type == 'measure'

        uom = uom_obj.browse (cr, uid, uom_id, context=context)

        if configurator.dimension_id.measure_uom_category and uom_id != comfigurator.dimension_id.measure_uom_category.id:
            raise osv.except_osv (_('Value error'),
                   _('Dimension [%(name)s] accepts only a UoM belongs to category [%(uom_category)s]. Please check!') % {
                            'name': configurator.dimension_id.name,
                            'uom_category': configurator.dimension_id.measure_uom_category.name})

        configurator.product_id.conf_validate()

        self.write (cr, uid, ids, {'uom_id': uom_id}, context=context)
        configurator.refresh_value()

        return {'value': {
                    'values_ids': [v.id for v in configurator.product_id.get_all_values()],
                }}

    def onchange_selected_product_id (self, cr, uid, ids, selected_product_id, context=None):
        context = context or self.get_user_context(cr, uid)
        product_obj = self.pool.get("product.product")

        assert ids and len(ids)==1

        configurator = self.browse (cr, uid, ids[0], context=context)

        configurator.product_id.conf_validate()

        self.write (cr, uid, ids, {'selected_product_id': selected_product_id}, context=context)
        configurator.refresh_value()

        return {'value': {
                    'values_ids': [v.id for v in configurator.product_id.get_all_values()],
                }}


    def onchange_current_selection_id (self, cr, uid, ids, current_selection_id, context=None):
        context = context or self.get_user_context(cr, uid)
        sv_obj = self.pool.get ('product.conf_selection_value')

        assert ids and len(ids)==1

        configurator = self.browse (cr, uid, ids[0], context=context)

        configurator.product_id.conf_validate()

        self.write (cr, uid, ids, {'current_selection_id': current_selection_id}, context=context)
        configurator.refresh_value()

        return {'value': {
                    'values_ids': [v.id for v in configurator.product_id.get_all_values()],
                }}


    def action_close (self, cr, uid, ids, context=None):
        context = context or {}

        assert ids and len(ids)==1

        refresh_value (cr, uid, ids, context=contex)

        return {'type':'ir.actions.act_window_close' }


configurator()


