# -*- coding: utf-8 -*-
##############################################################################
#
#    NUMA
#    Copyright (C) 2011 NUMA Extreme Systems (<http:www.numaes.com>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta
import time

from osv import fields, osv
from tools.translate import _
import decimal_precision as dp
import netsvc
import traceback

import pdb

class product_conf_dimension_initial (osv.osv):
    _name = "product.conf_dimension"

product_conf_dimension_initial()

class product_conf_selection_value(osv.osv):
    _name = "product.conf_selection_value"
    _rec_name = 'representation'

    _columns = {
        'dimension_id': fields.many2one ('product.conf_dimension', 'Dimension', required=True), 
        'value': fields.char ('Internal value representation', size=32, required=True),
        'representation': fields.text ('User representation of the value', required=True),
        'description': fields.text ('Value long description', required=True),
        'picture': fields.binary ('Picture', filters="*.png,*.jpg,*.bmp"),
    }

    _defaults = {

    }

product_conf_selection_value()

class product_conf_dimension(osv.osv):
    _inherit = "product.conf_dimension"
    _order = 'seq'

    _columns = {
        'product_id': fields.many2one ('product.product', 'Product'),
        'name': fields.char ('Name', size=64, required=True),
        'seq': fields.integer ('Sequence', required=True), 
        'description': fields.text ('Prompt description'),
        'is_mandatory': fields.selection ([
                            ('mandatory', 'Always a mandatory value'),
                            ('computed', 'Computed based on current configuration'),
                            ('optional', 'Always an optional value')], 'Mandatory status', required=True),

        'atom_type': fields.selection ([
                            ('number', 'A number'),
                            ('text', 'Some text'),
                            ('currency', 'A currency'),
                            ('date', 'A date'),
                            ('timestamp', 'A point in time'),
                            ('time', 'A time length'),
                            ('measure', 'A measure'),
                            ('product', 'A product'),
                            ('selection', 'A value from a list')], 'Atom type'),

        'min_number': fields.float ('Minimun number', digits=(16,2)),
        'max_number': fields.float ('Max number', digits=(16,2), help="If lower than minimun, then no check"),
        'text_mask': fields.text ('Text mask', help="Regular expression to check the input or nothing for no check"),
        'default_currency_id': fields.many2one ('res.currency', 'Currency'),
        'earliest_date': fields.date ('Earliest date', help="Earliest date to accept or nothing for no check"),
        'latest_date': fields.date ('Latest date', help="Latest date to accept or nothing for no check"),
        'earliest_point_in_time': fields.datetime ('Earliest PiT', help="Earliest Point in Time to accept or nothing for no check"),
        'latest_point_in_time': fields.datetime ('Latest PiT', help="Latest Point in Time to accept or nothing for no check"),
        'max_time_lenght': fields.time ('Max time duration', help="Maximun time duration or nothing for no check"),
        'measure_uom_category': fields.many2one ('product.uom.categ', 'UoM Category'),
        'product_domain': fields.text ('Product domain', help="Product domain Python expression to limit the range of possible products, or nothing for no limitation"),
        'selection_list_ids': fields.one2many ('product.conf_selection_value', 'dimension_id', 'Base selection list', help="Default selection list"),
        'validation_code': fields.text ('Validation code', help="Python code to validate the entered value"),
        'picture': fields.binary ('Picture', filters="*.png,*.jpg,*.bmp"),

    }

    _defaults = {

    }


product_conf_dimension()


class product_conf_value(osv.osv):
    _name = "product.conf_value"

    def _get_str_value (self, cr, uid, ids, name=None, arg=None, context=None):
        context = context or {}

        res = {}
        for v in self.browse(cr, uid, ids, context=context):
            if v.dimension_id:
                if v.atom_type == 'number':
                    res [v.id] = str(v.number)
                elif v.atom_type == 'text':
                    res [v.id] = v.text
                elif v.atom_type == 'currency':
                    res [v.id] = v.currency_id and v.currency_id.symbol or False
                elif v.atom_type == 'date':
                    res [v.id] = v.date
                elif v.atom_type == 'timestamp':
                    res [v.id] = v.timestamp
                elif v.atom_type == 'time':
                    res [v.id] = v.time
                elif v.atom_type == 'measure':
                    res [v.id] = '%d %s' % (v.number, v.uom_id and v.uom_id.name or '')
                elif v.atom_type == 'product':
                    res [v.id] = v.selected_product_id and v.selected_product_id.name_get()[0][1] or False
                elif v.atom_type == 'selection':
                    res [v.id] = v.current_selection_id and v.current_selection_id.representation or False
                else:
                    res [v.id] = None
            else:
                res [v.id] = None

        return res

    _columns = {
        'product_id': fields.many2one ('product.product', 'Product', required=True), 
        'dimension_id': fields.many2one ('product.conf_dimension', 'Dimension', required=True), 

        'number': fields.float ('As number', digits=(16,2)),
        'text': fields.text ('As text or selection'),
        'currency_id': fields.many2one ('res.currency', 'As currency'),
        'date': fields.date ('As date'),
        'timestamp': fields.datetime ('As Point in Time'),
        'time': fields.time ('As elapsed time'),
        'uom_id': fields.many2one ('product.uom', 'As UoM'),
        'selected_product_id': fields.many2one ('product.product', 'As product'),
        'current_selection_id': fields.many2one ('product.conf_selection_value', 'Current selection'),

        'str_value': fields.function (_get_str_value, method=True, type= "char", string="Current value"), 
        'atom_type': fields.related ('dimension_id', 'atom_type', type = "char", size=32, string="Atom type", readonly=True),
    }

    def get_value (self, cr, uid, ids, context=None):
        context = context or {}

        assert ids and len(ids)==1

        v = self.browse (cr, uid, ids[0], context=context)
        if v.dimension_id:
            if v.atom_type == 'number':
                return v.number
            elif v.atom_type == 'text':
                return v.text
            elif v.atom_type == 'currency':
                return v.currency_id
            elif v.atom_type == 'date':
                return v.date
            elif v.atom_type == 'timestamp':
                return v.timestamp
            elif v.atom_type == 'time':
                return v.time
            elif v.atom_type == 'measure':
                return {'number': v.number, 'uom_id': v.uom_id}
            elif v.atom_type == 'product':
                return v.selected_product_id
            elif v.atom_type == 'selection':
                return v.current_selection_id
        return None

    def set_value (self, cr, uid, ids, new_value, context=None):
        context = context or {}

        assert ids and len(ids)==1

        v = self.browse (cr, uid, ids[0], context=context)
        if v.dimension_id:
            if v.atom_type == 'number':
                vals = {'number': new_value}
            elif v.atom_type == 'text':
                vals = {'text': new_value}
            elif v.atom_type == 'currency':
                vals = {'currency_id': isinstance(new_value, int) and new_value or new_value.id}
            elif v.atom_type == 'date':
                vals = {'date': new_value}
            elif v.atom_type == 'timestamp':
                vals = {'timestamp': new_value}
            elif v.atom_type == 'time':
                vals = {'time': new_value}
            elif v.atom_type == 'measure':
                vals = {'number': new_value['number'], 'uom_id': isinstance(new_value['uom_id'], int) and new_value['uom_id'] or new_value['uom_id'].id}
            elif v.atom_type == 'product':
                vals = {'selected_product_id': isinstance(new_value, int) and new_value or new_value.id}
            elif v.atom_type == 'selection':
                vals = {'current_selection_id': isinstance(new_value, int) and new_value or new_value.id}
            else:
                return

            self.write (cr, uid, [v.id], vals, context=context)
    
        return

product_conf_value()

class product_conf_component(osv.osv):
    _name = "product.conf_component"

    _columns = {
        'name': fields.char ('Name', size=32),
        'label': fields.char ('Label', size=32),
        'component_id': fields.many2one ('product.product', 'Component product', required=True),
        'product_id': fields.many2one ('product.product', 'Container product', required=True),
        'comment': fields.text ('Optional comment'), 
        'quantity': fields.float ('Quantity', digits=(16,2), required=True),
        'uom_id': fields.related ('product_id', 'uom_id', string='UoM', type="many2one", obj="product.uom", readonly=True),
        'category_id': fields.related ('product_id', 'categ_id', string='Category', type="many2one", obj="product.category", readonly=True),
    }

    _defaults = {

    }

    def onchange_component_id (self, cr, uid, ids, component_id, context=None):
        context = context or {}

        product_obj = self.pool.get ('product.product')

        if component_id:
            product = product_obj.browse (cr, uid, component_id, context=context)

            res ={
                'category_id': product.categ_id.id,
                'uom_id': product.uom_id.id,
            }

            return {'value': res}

        return False

product_conf_component()

class product_conf_bom_rule_base (osv.osv):
    _name = "product.conf_bom_rule"

product_conf_bom_rule_base()

class product_conf_bom_dimension (osv.osv):
    _name = "product.conf_bom_dimension"

    _columns = {
        'bom_rule_id': fields.many2one ('product.conf_bom_rule', 'BoM Rule'),
        'name': fields.char ('Name', size=32),
        'product_id': fields.related ('bom_rule_id', 'product_id', string='Product', type="many2one", obj='product.product', readonly=True),
        'dimension_domain': fields.related ('product_id', 'conf_all_dimensions_ids', string='Valid dimensions for product', type="one2many", obj="product.conf_dimension", readonly="1"),

        'dimension_source': fields.selection ([
                            ('fixed', 'Fixed dimension'),
                            ('dimension', 'Dimension value'),
                            ('code', 'Python code')], 'Dimension source', required=True),
        'dimension_name': fields.char ('Dimension name', size = 64),
        'dimension_dimension_id': fields.many2one ('product.conf_dimension', 'Dimension to use', domain="[('id','in',product_dimensions_ids)]"),
        'dimension_code': fields.text ('Dimension selection code'),

        'dimension_atom_type': fields.related ('dimension_dimension_id', 'atom_type', type="char", size=32, string="Atom type", readonly=True),
        'dimension_selection_domain': fields.related ('dimension_dimension_id', 'selection_list_ids', type="one2many", obj="product.conf_selection_value", string="Selection domain"),

        'value_source': fields.selection ([
                            ('fixed', 'Fixed value'),
                            ('dimension', 'Get value from Dimension'),
                            ('code', 'Python code')], 'Dimension value source', required=True),

        'number': fields.float ('As number', digits=(16,2)),
        'text': fields.text ('As text or selection'),
        'currency_id': fields.many2one ('res.currency', 'As currency'),
        'date': fields.date ('As date'),
        'timestamp': fields.datetime ('As Point in Time'),
        'time': fields.time ('As elapsed time'),
        'uom_id': fields.many2one ('product.uom', 'As UoM'),
        'selected_product_id': fields.many2one ('product.product', 'As product'),
        'current_selection_id': fields.many2one ('product.conf_selection_value', 'Current selection'),

        'value_dimension_id': fields.many2one ('product.conf_dimension', 'Dimension to use', domain="[('id','in',product_dimensions_ids)]"),
        'value_code': fields.text ('Value selection code'),
    }

    _defaults = {
        'dimension_source': 'fixed',
        'value_source': 'fixed',
        'dimension_code': """
# Input parameters: 
#   product         (container product)
#   dimension       (initially set to False)
#   pool
#   cr
#   uid
#
# Output expected:
#   dimension       (browse record)   

""",
        'value_code': """
# Input parameters: 
#   product         (container product)
#   dimension       (browse record)   
#   value           (initially set to False)   
#   pool
#   cr
#   uid
#
# Output expected:
#   value           (proper value for dimension)   

""",
    }

    def onchange_dimension_source (self, cr, uid, ids, dimension_source, context=None):
        context = context or {}

        if dimension_source != 'fixed':
            return {'value': {
                        'dimension_atom_type': False,
                        'dimension_dimension_id': False, 
                        'dimension_selection_domain': [],
                        }}

        return False

    def onchange_dimension_dimension_id (self, cr, uid, ids, dimension_dimension_id, context=None):
        context = context or {}
        dimension_obj = self.pool.get('product.conf_dimension')

        if dimension_dimension_id:
            return {'value': {
                        'dimension_selection_domain': dimension_obj.browse(cr, uid, dimension_dimension_id, context=context).selection_list_ids,
                        }}

        return False


    def get_dimension (self, cr, uid, ids, target_product, context=None):
        context = context or {}

        assert ids and len(ids)==1

        bd = self.browse (cr, uid, ids[0], context=context)

        res = {}

        if bd.dimension_source == 'fixed':
            res['dimension'] = target_product.get_conf_dimension(bd.dimension_name).id
        elif bd.dimension_source == 'dimension':
            value = target_product.get_conf_value(bd.dimension_dimension_id.name)
            res['dimension'] = value.get_value()
        else:
            if not bd.dimension_code:
                raise osv.except_osv (_('Configuration problem'),
                       _('BoM dimension rule wrongly configured [%(name)s], rule sequence %(sequence)d on product [%(product)s] dimension set to code computation, but no code is provided. Please check!') % {
                                    'name': bd.name, 
                                    'sequence': bd.bom_rule_id.seq, 
                                    'product': bd.product_id.name})
            compute_context = {
                'product': product,
                'dimension': None,
                'pool': self.pool,
                'cr': cr,
                'uid': uid
            }

            try:
                compute_context.update(context)
                exec bd.dimension_code in compute_context
                res['dimension'] = compute_context['dimension'].id
            except Exception, e:
                raise osv.except_osv (_('Configuration problem'),
                       _('BoM dimension rule wrongly configured [%(name)s], rule sequence %(sequence)d on product [%(product)s] dimension set to code, computation has thrown an unexpected exception [%(exception)s]. Please check!') % {
                                    'name': bd.name, 
                                    'sequence': bd.bom_rule_id.seq, 
                                    'product': bd.product_id.name,
                                    'exception': e})

        if bd.value_source == 'fixed':
            if bd.dimension_atom_type:
                if bd.dimension_atom_type == 'number':
                    res['value'] = bd.number
                elif bd.dimension_atom_type == 'text':
                    res['value'] = bd.text
                elif bd.dimension_atom_type == 'currency':
                    res['value'] = bd.currency_id
                elif bd.dimension_atom_type == 'date':
                    res['value'] = bd.date
                elif bd.dimension_atom_type == 'timestamp':
                    res['value'] = bd.timestamp
                elif bd.dimension_atom_type == 'time':
                    res['value'] = bd.time
                elif bd.dimension_atom_type == 'measure':
                    res['value'] = {'number': bd.number, 'uom_id': bd.uom_id}
                elif bd.dimension_atom_type == 'product':
                    res['value'] = bd.selected_product_id
                elif bd.dimension_atom_type == 'selection':
                    res['value'] = bd.current_selection_id
            else:
                res['value'] = None
        elif bd.value_source == 'dimension':
            value = target_product.get_conf_value(bd.dimension_dimension_id.name)
            res['value'] = value
        else:
            if not bd.value_code:
                raise osv.except_osv (_('Configuration problem'),
                       _('BoM dimension rule wrongly configured [%(name)s], rule sequence %(sequence)d on product [%(product)s] value set to code computation, but no code is provided. Please check!') % {
                                    'name': bd.name, 
                                    'sequence': bd.bom_rule_id.seq, 
                                    'product': bd.product_id.name})
            compute_context = {
                'product': product,
                'dimension': res['dimension'],
                'value': None,
                'pool': self.pool,
                'cr': cr,
                'uid': uid
            }

            try:
                compute_context.update(context)
                exec bd.value_code in compute_context
                res['value'] = compute_context['value']
            except Exception, e:
                raise osv.except_osv (_('Configuration problem'),
                       _('BoM dimension rule wrongly configured [%(name)s], rule sequence %(sequence)d on product [%(product)s] dimension set to code, computation has thrown an unexpected exception [%(exception)s]. Please check!') % {
                                    'name': bd.name, 
                                    'sequence': bd.bom_rule_id.seq, 
                                    'product': bd.product_id.name,
                                    'exception': e})

        return res

product_conf_bom_dimension()    

class product_conf_bom_rule(osv.osv):
    _inherit = "product.conf_bom_rule"
    _order = 'seq'

    _columns = {
        'product_id': fields.many2one ('product.product', 'Product', required=True),
        'dimension_domain': fields.related ('product_id', 'conf_all_dimensions_ids', string='Valid dimensions for product', type="one2many", obj="product.conf_dimension"),
        'seq': fields.integer('Sequence', required=True),
        'name': fields.char ('Nombre', size=32, required=True),

        'product_source': fields.selection ([
                            ('fixed', 'Fixed product'),
                            ('dimension', 'Dimension value'),
                            ('code', 'Python code')], 'Product source', required=True),
        'product_selection_id': fields.many2one ('product.product', 'Selected product'),
        'product_dimension_id': fields.many2one ('product.conf_dimension', 'Dimension to use', domain="[('id','in',product_dimensions_ids)]"),
        'product_code': fields.text ('Product selection code'),

        'quantity_source': fields.selection ([
                            ('fixed', 'Fixed quantity'),
                            ('dimension', 'Dimension value'),
                            ('code', 'Python code')], 'Quantity source', required=True),
        'quantity_selection': fields.float ('Fixed quantity', digits=(16,2)),
        'quantity_dimension_id': fields.many2one ('product.conf_dimension', 'Dimension to use', domain="[('id','in',product_dimensions_ids)]"),
        'quantity_code': fields.text ('Quantity selection code'),

        'conf_bom_dimensions_ids': fields.one2many ('product.conf_bom_dimension', 'bom_rule_id', 'Dimensions to configure'),
        'bom_dimensions_code': fields.text ('BoM dimensions selection code'),
 
    }

    _defaults = {
        'product_source': 'fixed',
        'quantity_source': 'fixed',
        'product_code': """
# Input parameters: 
#   product         (container product)
#   component       (initially set to False)
#   pool
#   cr
#   uid
#
# Output expected:
#   component       (browse record)

""",
        'quantity_code': """
# Input parameters: 
#   product         (container product)
#   component       (browse record)
#   quantity        (initially set to 1.0)
#   pool
#   cr
#   uid
#
# Output expected:
#   quantity

""",
        'bom_dimensions_code': """
# Input parameters: 
#   product         (container product)
#   component       (browse record)
#   quantity        (From previous step)
#   dimensions  A dictionary with keys:
#                   dimension       (browse record)
#                   value
#   pool
#   cr
#   uid
#
# Output expected:
#   dimensions      (dictionary, same keys as input parameter)

""",
    }

    def get_product_and_qty (self, cr, uid, ids, target_product, context=None):
        conf_dimension_value_obj = self.pool.get ('product.conf_dimension_value')

        assert ids and len(ids)==1

        res = {}
        conf_rule = self.browse (cr, uid, ids[0], context=context)

        res['name'] = conf_rule.name

        product = conf_rule.product_id
        if conf_rule.product_source == 'fixed':
            if not conf_rule.product_selection_id:
                raise osv.except_osv (_('Configuration problem'),
                       _('BoM Rule wrongly configured. Rule sequence %(sequence)d on product [%(product)s] set to fixed product but no product is configured. Please check!') % {
                                    'sequence': conf_rule.seq, 
                                    'product': product.name})
            res['component_id'] = conf_rule.product_selection_id.id
        elif conf_rule.product_source == 'dimension':
            value = target_product.get_conf_value(conf_rule.product_dimension_id.name)
            if not value:
                raise osv.except_osv (_('Configuration problem'),
                       _('BoM Rule wrongly configured. Rule sequence %(sequence)d on product [%(product)s] product set to dimension value, but it was not found or it has multiple values. Please check!') % {
                                    'sequence': conf_rule.seq, 
                                    'product': product.name})
            res['component_id'] = value.get_value().id
        else:
            if not conf_rule.product_code:
                raise osv.except_osv (_('Configuration problem'),
                       _('BoM Rule wrongly configured. Rule sequence %(sequence)d on product [%(product)s] product set to code computation, but no code is provided. Please check!') % {
                                    'sequence': conf_rule.seq, 
                                    'product': product.name})
            compute_context = {
                'product': product,
                'component': False,
                'pool': self.pool,
                'cr': cr,
                'uid': uid
            }

            try:
                compute_context.update(context)
                exec conf_rule.product_code in compute_context
                if compute_context['component_id']:
                    res['component_id'] = compute_context['component'].id
                else:
                    raise osv.except_osv (_('Configuration problem'),
                           _('BoM Rule code on sequence %(sequence)d, product [%(product)s] has return no component_id to configure. Please check!') % {
                                        'sequence': conf_rule.seq, 
                                        'product': product.name})

            except Exception, e:
                raise osv.except_osv (_('Configuration problem'),
                       _('BoM Rule code on sequence %(sequence)d, product [%(product)s] has thrown an unexpected exception [%(exception)s]. Please check!') % {
                                    'sequence': conf_rule.seq, 
                                    'product': product.name, 
                                    'exception': e})

        if conf_rule.quantity_source == 'fixed':
            res['quantity'] = conf_rule.quantity_selection
        elif conf_rule.quantity_source == 'dimension':
            value = target_product.get_conf_value(conf_rule.quantity_dimension_id.name)
            if not value:
                raise osv.except_osv (_('Configuration problem'),
                       _('BoM Rule wrongly configured. Rule sequence %d on product [%s], quantity set to dimension value, but it was not found or it has multiple values. Please check!') % {
                                    'sequence': conf_rule.seq, 
                                    'product': product.name})
            res['quantity'] = float(value)
        else:
            if not conf_rule.quantity_code:
                raise osv.except_osv (_('Configuration problem'),
                       _('BoM Rule wrongly configured. Rule sequence %(sequence)d on product [%(product)s] quantity set to code computation, but no code is provided. Please check!') % {
                                    'sequence': conf_rule.seq, 
                                    'product': product.name})
            compute_context = {
                'product': product,
                'component': res['component_id'],
                'quantity': 1.0,
                'pool': self.pool,
                'cr': cr,
                'uid': uid
            }

            try:
                compute_context.update(context)
                exec conf_rule.quantity_code in compute_context
                res['quantity'] = compute_context['quantity']
            except Exception, e:
                raise osv.except_osv (_('Configuration problem'),
                       _('BoM Rule code on sequence %(sequence)d, product [%(product)s]. Quantity code has thrown an unexpected exception [%(exception)s]. Please check!') % {
                                    'sequence': conf_rule.seq, 
                                    'product': product.name, 
                                    'exception': e})

        # Compute dimensions
        dims = []
        for cd in conf_rule.conf_bom_dimensions_ids:
            dims.append (cd.get_dimension(product))

        if conf_rule.bom_dimensions_code:
            compute_context = {
                'product': product,
                'component': res['component_id'],
                'quantity': res['quantity'],
                'dimensions': dims,
                'pool': self.pool,
                'cr': cr,
                'uid': uid
            }

            try:
                compute_context.update(context)
                exec conf_rule.bom_dimensions_code in compute_context
                res ['dimensions'] = compute_context['dimensions']
            except Exception, e:
                raise osv.except_osv (_('Configuration problem'),
                       _('BoM Rule dimensions code on sequence %(sequence)d, product [%(product)s]. Dimensions code has thrown an unexpected exception [%(exception)s]. Please check!') % {
                                    'sequence': conf_rule.seq, 
                                    'product': product.name, 
                                    'exception': e})
        else:
            res ['dimensions'] = dims

        return res        

    _defaults = {

    }

product_conf_bom_rule()

def _user_exception (msg):
    return osv.except_osv (_('Configuration problem'), msg)

class product_product(osv.osv):
    _inherit = "product.product"

    def _get_all_dimensions (self, cr, uid, ids, field_name, args, context=None):
        context = context or {}

        res = {}
        for id in ids:
            res[id] = self.get_all_conf_dimensions_ids (cr, uid, [id], context=context)

        return res
  
    def _get_all_values (self, cr, uid, ids, field_name, args, context=None):
        context = context or {}

        res = {}
        for id in ids:
            res[id] = self.get_all_conf_values_ids (cr, uid, [id], context=context)

        return res

    _columns = {
        'conf_dimensions': fields.one2many('product.conf_dimension', 'product_id', 'Own configuration dimensions'),
        'conf_all_dimensions_ids': fields.function (_get_all_dimensions, method=True, type="one2many", obj="product.conf_dimension", string="All dimensions"),
        'conf_values': fields.one2many('product.conf_value', 'product_id', 'Own configuration values'),
        'conf_all_values_ids': fields.function (_get_all_values, method=True, type="one2many", obj="product.conf_value", string="All values"), 
        'conf_parent_ids': fields.many2many('product.product', 'product_conf_parents', 'product', 'parent', string='Configuration parent'),

        'conf_use_parent_bom': fields.boolean ('Use parent BoM?', help="Use the compute BoM from parent (and eventually all the chain of parents) as base for the current BoM (in addition to the own BoM generated on BoM rules and code)"),
        'conf_bom_rules_ids': fields.one2many ('product.conf_bom_rule', 'product_id', 'Rules of BOM creation'),
        'conf_bom_ids': fields.one2many('product.conf_component', 'product_id', string='BoM Components'),
        'conf_bom_creation_code': fields.text ('BoM creation code'),

        'conf_mask_code': fields.text ('Mask code', help="Dimension and selection lists masking code"),
        'conf_validate_code': fields.text ('Configuration validation code'),

    }

    _defaults = {
        'conf_use_parent_bom': True,
        'conf_bom_creation_code': """
# Input parameters: 
#   product             (browse record, container product)
#   new_bom             dictionary with bom computed from plus the BoM rules
#   values              all dimension values of the product (own and inherited) (List of browse records)
#   dimensions          all dimensions of the product (own and inherited) (List of browse records)
#   product_obj         utility. product_obj
#   component_obj       utility. component_obj
#   user_exception      utility method to generate user exceptions. It accepts a single string parameter
#   dimensions          List of dimensions (browse records)
#   dimensions_mask     A dictionary with keys=dimensions.id and a boolean value indicating mandatory dimension
#   pool
#   cr
#   uid
#
# Output expected:
#   dimensions_mask     (dictionary, same keys as input parameter)

""",
        'conf_mask_code': """
# Input parameters: 
#   product             (browse record, container product)
#   user_exception      utility method to generate user exceptions. It accepts a single string parameter
#   dimensions          List of dimensions (browse records)
#   dimensions_mask     A dictionary with keys=dimensions.id and a boolean value indicating mandatory dimension
#   pool
#   cr
#   uid
#
# Output expected:
#   dimensions_mask     (dictionary, same keys as input parameter)

""",
        'conf_validate_code': """
# Input parameters: 
#   product             (browse record, container product)
#   user_exception      utility method to generate user exceptions. It accepts a single string parameter
#   pool
#   cr
#   uid
#
# No output is expected. If there is any errors, a user exception should be thrown using the user_exception
# utility method

""",
    }

    def _check_conf_parent_recursion (self, cr, uid, parent, product, context=None):
        context=context or {}

        for parent in self.browse (cr, uid, parent.conf_parent_ids, context=context):
            if parent == product:
                raise osv.except_osv (_('Configuration problem'),
                       _('RECURSION DETECTED. New configuration parent would create a circular reference, and thus it is not allowed. Please check!'))
            self._check_conf_parent_recursion (cr, uid, parent, product, context=context)

        return

    def onchange_conf_parent_ids (self, cr, uid, ids, parent_ids, context=None):
        context = context or {}

        assert ids and len(ids)==1

        if parent_ids:
            product = self.browse (cr, uid, ids[0], context=context)
            # Check recursion
            p_ids = parent_ids[0][2]
            for parent in self.browse (cr, uid, p_ids, context=context):
                if parent == product:
                    raise osv.except_osv (_('Configuration problem'),
                           _('RECURSION DETECTED. New configuration parent would create a circular reference, and thus it is not allowed. Please check!'))
                self._check_conf_parent_recursion (cr, uid, parent, product, context=context)

        return False

    def conf_mask (self, cr, uid, ids, context=None):
        context = context or {}

        assert ids and len(ids)==1

        product = self.browse (cr, uid, ids[0], context=context)
        dimensions = self.get_all_dimensions (cr, uid, [product.id], context=context)
        dimensions_mask = {}
        for d in dimensions:
            selection_mask = {}
            for s in d.selection_list_ids:
                selection_mask[s.value] = True
            dimensions_mask[d.id] = {'mandatory': d.is_mandatory == 'mandatory', 'selection_mask': selection_mask}

        compute_context = {
            'product': product,
            'user_exception': _user_exception, 
            'dimensions': dimensions,
            'dimensions_mask': dimensions_mask,
            'uid': uid,
            'cr': cr,
            'pool': self.pool,
        }
        compute_context.update(context)
        if product.conf_validate_code:
            exec product.conf_validate_code in compute_context
        return (compute_context['dimensions'], compute_context['dimensions_mask'])

    def conf_validate (self, cr, uid, ids, context=None):
        context = context or {}

        assert ids and len(ids)==1

        product = self.browse (cr, uid, ids[0], context=context)
        compute_context = {
            'product': product,
            'user_exception': _user_exception, 
            'uid': uid,
            'cr': cr,
            'pool': self.pool,
        }
        try:
            compute_context.update(context)
            exec product.conf_validate_code in compute_context
        except osf.except_osv, e:
            raise e
        except Exception, e:
            raise osv.except_osv (_('Configuration problem'),
                   _('Validate code in product [%(product)s] has thrown an unexpected exception [%(exception)s]. Please check!') % {
                            'product': product.name,
                            'exception': e})
        return

    def get_conf_dimension (self, cr, uid, ids, dimension_name, context=None):
        context = context or {}
        dimension_obj = self.pool.get('product.conf_dimension')

        assert ids and len(ids)==1

        dim_ids = dimension_obj.search (cr, uid, [('name','=', dimension_name), ('id', 'in', self.get_all_conf_dimensions_ids (cr, uid, ids, context=context))], context=context)

        if not dim_ids:
            raise osv.except_osv (_('Configuration problem'),
                   _('Dimension [%(name)s] not found in product [%(product)s]. Please check!') % {
                            'name': dimension_name,
                            'product': product.name,
                            'exception': e})

        return dimension_obj.browse (cr, uid, dim_ids[0], context=contex)

    def get_all_conf_dimensions_ids (self, cr, uid, ids, context=None):
        context = context or {}

        assert ids and len(ids)==1

        res = {}

        product = self.browse (cr, uid, ids[0], context=context)

        if not context.get('_recursion_check', False):
            recursion_check = set()
            context = context.copy()
        else:
            recursion_check = context['_recursion_check']

        if product.id in recursion_check:
            raise osv.except_osv (_('Configuration problem'),
                   _('RECURSION DETECTED on all dimensions get of product [%(product)s]. Please check configuration parents for circular references!') % {'product':product.name})
        recursion_check.add(product.id)

        res = product.conf_dimensions and [d.id for d in product.conf_dimensions] or []

        context['_recursion_check'] = recursion_check

        if product.conf_parent_ids:
            #Set used to check if a parent was already visited.
            #Parents should add dimensions only once
            parent_set = context.get('_parent_set', False) or set()
   
            for parent in product.conf_parent_ids:
                if parent not in parent_set:
                    parent_set.add(parent)
                    context['_parent_set'] = parent_set
                    res += self.get_all_conf_dimensions_ids(cr, uid, [parent.id], context=context) or []

        res.sort()

        return res

    def get_all_dimensions (self, cr, uid, ids, context=None):
        context = context or {}

        assert ids and len(ids)==1

        dimension_obj = self.pool.get ('product.conf_dimension')

        dim_ids = self.get_all_conf_dimensions_ids (cr, uid, ids, context=context)
        dims = dim_ids and dimension_obj.browse (cr, uid, dim_ids, context=context) or []
        dims = sorted(dims, key = lambda dim: dim.seq)
        return dims

    def get_all_conf_values_ids (self, cr, uid, ids, context=None):
        context = context or {}

        assert ids and len(ids)==1

        res = {}
        recorded_dims = {}

        product = self.browse (cr, uid, ids[0], context=context)

        if not context.get('_recursion_check', False):
            recursion_check = set()
            context = context.copy()
        else:
            recursion_check = context['_recursion_check']

        if product.id in recursion_check:
            raise osv.except_osv (_('Configuration problem'),
                   _('RECURSION DETECTED on all values get of product [%(product)s]. Please check configuration parents for circular references!') % {'product':product.name})
        recursion_check.add(product.id)

        new_values_ids = product.conf_values and [v.id for v in product.conf_values] or []
        new_values_dims_ids = product.conf_values and [v.dimension_id.id for v in product.conf_values] or []
        for vd_id in new_values_dims_ids:
            recorded_dims[vd_id] = True
        res = new_values_ids

        if product.conf_parent_ids:
            context['_recursion_check'] = recursion_check

            #Set used to check if a parent was already visited.
            #Parents should add dimensions only once
            parent_set = context.get('_parent_set', False) or set()
   
            new_values = []
            for parent in product.conf_parent_ids:
                if parent not in parent_set:
                    parent_set.add(parent)
                    context['_parent_set'] = parent_set
                    new_values += self.get_all_values(cr, uid, [parent.id], context=context) or []

            new_values = filter (lambda v: v.dimension_id.id not in recorded_dims, new_values)
            new_values_ids = []
            for v in new_values:
                new_values_ids.append(v.id)
            res += new_values_ids
        res.sort()

        return res

    def get_all_values (self, cr, uid, ids, context=None):
        context = context or {}

        assert ids and len(ids)==1

        value_obj = self.pool.get ('product.conf_value')

        product = self.browse (cr, uid, ids[0], context=context)
        val_ids = self.get_all_conf_values_ids (cr, uid, [product.id], context=context)
        vals = val_ids and value_obj.browse (cr, uid, val_ids, context=context) or []
        return sorted(vals, key = lambda val: val.dimension_id.seq)

    def is_configurable (self, cr, uid, ids, context=None):
        context = context or {}

        assert ids and len(ids)==1

        res = {}
        product = self.browse (cr, uid, ids[0], context=context)
        if product.conf_dimensions:
            return True
        else:
            if product.conf_parent_ids:
                for parent in product.conf_parent_ids:
                    if self.is_configurable(cr, uid, [parent.id], context=context):
                        return True

        return False

    def get_product_from_code (self, cr, uid, ids, code, search_values={}, context=None):
        context = context or {}

        product_ids = self.search (cr, uid, [('code','=', code)], context=contex)
        if not product_ids:
            raise osv.except_osv (_('Configuration problem'),
                   _('Product code not found [%(code)s] in get_product_from_code. Please check!') % {
                            'product': code})

        products = self.browse (cr, uid, product_ids, context=context)

        if values:
            filtered_products = []
            for product in products:
                for value in product.get_all_values():
                    if (value.dimension_id.name in search_values) and (search_values[value.dimension_id.name] == value.value()):
                        filtered_products.append(product)
            return filtered_products
        else:
            return products

    def _create_bom (self, cr, uid, product, target, context=None):
        context = context or {}
        value_obj = self.pool.get('product.conf_value')
        dimension_obj = self.pool.get('product.conf_dimension')    

        #Product is the static source, target the real object to process

        if product.conf_use_parent_bom and product.conf_parent_ids:
            new_bom = []
            parent_set = context.get('_parent_set', False) or set()
            for parent in product.conf_parent_ids:
                if parent not in parent_set:
                    parent_set.add(parent)
                    context['_parent_set'] = parent_set
                    new_bom += self._create_bom (cr, uid, parent, target, context=context)
        else:
            new_bom = []

        for rule in product.conf_bom_rules_ids:
            new_component = rule.get_product_and_qty(target, context=context)
            new_component['label'] = product.conf_label 
            new_bom.append(new_component)

        if product.conf_bom_creation_code:
            compute_context = {
                'product': target,
                'user_exception': _user_exception, 
                'new_bom': new_bom,
                'values': product.get_all_values (context=context),
                'dimensions': product.get_all_dimensions(context=context),
                'product_obj': self,
                'component_obj': self.pool.get('product.conf_component'),
                'uid': uid,
                'cr': cr,
                'pool' : self.pool,
            }
            try:
                compute_context.update(context)
                exec product.conf_bom_creation_code in compute_context
                new_bom = compute_context['new_bom']
            except Exception, e:
                raise osv.except_osv (_('Configuration problem'),
                       _('Dimensions get code in product [%(product)s], target [%(target)s] has thrown an unexpected exception [%(exception)s]. Please check!') % {
                                'product': product.name,
                                'target': target.name,
                                'exception': e})

        return new_bom

    def action_create_subproduct (self, cr, uid, ids, context=None):
        if context is None: context = {}

        assert ids and len(ids)==1

        product = self.browse (cr, uid, ids[0], context=context)

        vals = {
            'product_tmpl_id': product.product_tmpl_id.id,
            'conf_parent_id': [(6, 0, [product.id])],
            'name' : product.name, 
            'default_code' : "%s - Subproduct" % product.default_code,
        }

        sp_instance_id = self.create(cr, uid, vals, context=dict(context, active_ids=ids))

        return {
            'name':_("Create subproduct"),
            'view_mode': 'form',
            'view_id': False,
            'view_type': 'form',
            'res_model': 'product.product',
            'res_id': sp_instance_id,
            'type': 'ir.actions.act_window',
            'nodestroy': True,
            'domain': '[]',
            'context': dict(context, active_ids=ids)
        }

    def action_create_bom (self, cr, uid, ids, context=None):
        context = context or {}

        for product in self.browse(cr, uid, ids, context=context):

            if not product.is_configurable():
                continue

            to_delete = []
            for conf in product.conf_bom_ids:
                to_delete.append((2, conf.id))

            #remove existing boms
            self.write (cr, uid, [product.id], {'conf_bom_ids': to_delete}, context=context)

            new_bom = self._create_bom (cr, uid, product, product, context=context)
            new_values = [(0, 0, v) for v in new_bom]
            self.write (cr, uid, [product.id], {'conf_bom_ids': new_values}, context=context)

        return False

    def set_conf_value (self, cr, uid, ids, dimension_name, new_value, context=None):
        context = context or {}
        dimension_obj = self.pool.get ('product.conf_dimension')
        value_obj = self.pool.get ('product.conf_value')

        for product in self.browse(cr, uid, ids, context=context):
            dimension_ids = get_all_dimensions_ids (cr, uid, [product.id], context=context) [product.id]
            dimension_ids = dimension_obj.search (cr, uid, [('name','=',dimension_name),('id','in',dimension_ids)], context=context)
            if dimension_ids and len(dimension_ids)>1 or not dimension_ids:
                raise osv.except_osv (_('Configuration problem'),
                       _('Invalid dimension name [%(dimension)s] for product [%(product)s]. Value cannot be set. Please check!') % {'product':product.name, 'dimension':dimension_name})

            value_ids = self.get_all_value_ids (cr, uid, [product.id], context=context)[product.id]
            value_ids = value_obj.search (cr, uid, [('dimension_id','in',dimension_ids),('id','in', value_ids)], context=context)
            if not value_ids:
                value_id = value_obj.create (cr, uid, {'product_id': product.id, 'dimension_id': dimension_ids[0]}, context=context)
            else:
                value_id = value_ids[0]

            value = value_obj.browse (cr, uid, value_id, context=context)
            value.set_value(new_value, context=context)
        return        

    def get_conf_value (self, cr, uid, ids, dimension_name, context=None):
        context = context or {}
        dimension_obj = self.pool.get ('product.conf_dimension')
        value_obj = self.pool.get ('product.conf_value')
        uom_obj = self.pool.get ('product.uom')
        currency_obj = self.pool.get ('res.currency')

        assert ids and len(ids) == 1

        for product in self.browse(cr, uid, ids, context=context):
            dimension_ids = self.get_all_conf_dimensions_ids (cr, uid, [product.id], context=context)
            dimension_ids = dimension_obj.search (cr, uid, [('name','=',dimension_name),('id','in',dimension_ids)], context=context)
            if dimension_ids and len(dimension_ids)>1 or not dimension_ids:
                raise osv.except_osv (_('Configuration problem'),
                       _('Invalid dimension name [%(dimension)s] for product [%(product)s]. Value cannot be set. Please check!') % {'product':product.name, 'dimension':dimension_name})

            value_ids = self.get_all_conf_values_ids (cr, uid, [product.id], context=context)
            value_ids = value_obj.search (cr, uid, [('dimension_id','in',dimension_ids),('id','in', value_ids)], context=context)
            if not value_ids:
                 return None

            assert len(value_ids)==1

            value = value_obj.browse (cr, uid, value_ids[0], context=context)
            return value.get_value()

    def name_get (self, cr, uid, ids, context=None):
        context = context or {}

        res = []
        for rid, name in super(product_product, self).name_get(cr, uid, ids, context=context):
            product = self.browse (cr, uid, rid, context=context)
            value_descriptions = []
            for v in product.get_all_values():
                value_descriptions.append("%s %s" % (v.dimension_id.name, v.str_value))
            if len(value_descriptions)>0:
                name += ': '
                name += ', '.join (value_descriptions)
            res.append ((rid, name))

        return res

    def price_get(self, cr, uid, ids, ptype='list_price', context=None):
        context = context or {}

        res = {}

        if not context.get('_recursion_check', False):
            recursion_check = set()
            context = context.copy()
        else:
            recursion_check = context['_recursion_check']

        for product in self.browse (cr, uid, ids, context=context):
            if product in recursion_check:
                raise osv.except_osv (_('Configuration problem'),
                       _('RECURSION DETECTED on price computation of product [%(product)s]. Please check components in BoMs for circular references!') % {'product':product.name})
            recursion_check.add(product)
                
            if product.conf_bom_ids:
                # It is a compound product. Use BoM to compute price
                acumulated_price = 0.0
                for component in product.conf_bom_ids:
                    new_context = context.copy()
                    new_context['quantity'] = component.quantity
                    new_context['_recursion_check'] = recursion_check
                    component_unit_price = self.price_get (cr, uid, [component.component_id.id], ptype, context=new_context) [component.component_id.id]
                    acumulated_price += component_unit_price * component.quantity

                res[product.id] = acumulated_price
            else:
                res[product.id] = super (product_product, self).price_get (cr, uid, [product.id], ptype, context=context)[product.id]

        return res

    def action_configure (self, cr, uid, ids, context=None):
        if context is None: context = {}
        conf_instance_id = self.pool.get("product.conf_wizard").create(
            cr, uid, {'product_id': ids[0]}, context=dict(context, active_ids=ids))
        return {
            'name':_("Products to configure"),
            'view_mode': 'form',
            'view_id': False,
            'view_type': 'form',
            'res_model': 'product.conf_wizard',
            'res_id': conf_instance_id,
            'type': 'ir.actions.act_window',
            'nodestroy': True,
            'target': 'new',
            'domain': '[]',
            'context': dict(context, active_ids=ids)
        }

product_product()


