# -*- coding: utf-8 -*-
##############################################################################
#
#    NUMA
#    Copyright (C) 2011 NUMA Extreme Systems (<http:www.numaes.com>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta
import time

from osv import fields, osv
from tools.translate import _
import decimal_precision as dp
import netsvc
import traceback

import pdb


class sales_order (osv.osv):
    _inherit = "sale.order"

    _columns = {
        'costlist_id': fields.many2one('product.pricelist', 'Pricelist', required=True, readonly=True, states={'draft': [('readonly', False)]}, help="Pricelist for current sales order."),
    }

sales_order ()

class sales_flat_bom_line (osv.osv):
    _inherit = "sale.sales_flat_bom_line"

    def _get_margin_percentage (self, cr, uid, ids, name=None, args=None, context=None):
        context = context or {}

        res = {}
        for sbm in self.browse (cr, uid, ids, context=context):
            if sbm.unit_cost:
                res[sbm.id] = ((sbm.unit_price - sbm.unit-cost) / sbm.unit_cost) * 100.0
            else:
                res[sbm.id] = 0.00

        return res

    _columns = {
        'order_line_id': fields.many2one ('sale.order.line', 'Sales order line'),
        'product_id': fields.many2one ('product.product', 'Product', required=True),
        'category_id': fields.related ('product_id', 'categ_id', string='Category', type="many2one", obj="product.categ", readonly=True),
        'unit_price': fields.float ('Unit Price', digits_compute= dp.get_precision('Sale Price')),
        'margin': fields.float('Margin', digits_compute= dp.get_precision('Sale Price')),
        'margin_percentage': fields.function (_get_margin_percentage, 'Margin [%]'),
        'unit_cost': fields.float ('Unit Cost', digits_compute= dp.get_precision('Sale Price')),
    }

    def onchange_unit_price (self, cr, uid, ids, unit_price, unit_cost, context=None):
        context = context or {}

        res={}
        res['margin'] = unit_price-unit_cost
        if unit_price:
            res['margin_percentage'] = ((unit_price-unit_cost)/unit_price) * 100.0
        else:
            res['margin_percentage'] = 0.00

        return {'value': res}

    def onchange_unit_cost (self, cr, uid, ids, unit_cost, unit_price, context=None):
        return self.onchange_unit_price (cr, uid, ids, unit_price, unit_cost, context=context)

    def onchange_margin (self, cr, uid, ids, margin, unit_cost, context=None):

        unit_price = unit_cost+margin
        res={}
        res['unit_price'] = unit_cost + margin

        if res['unit_price']:
            res['margin_percentage'] = (margin/res['unit_price']) * 100.0
        else:
            res['margin_percentage'] = 0.00

        return {'value': res}

    def onchange_margin_percentage (self, cr, uid, ids, margin_percentage, unit_cost, context=None):

        unit_price = unit_cost / (1 - (margin_percentage / 100.0))
        margin = unit_price - unit_cost
        res['margin'] = margin
        res['unit_price'] = unit_price

        return {'value': res}

sales_bom_line ()

class sales_order_line (osv.osv):
    _inherit = "sale.order.line"

    _columns = {
        'flat_bom_lines_ids': fields.one2many ('sale.sales_flat_bom_line', 'order_line_id', 'Flat BoM lines'),
        'unit_cost': fields.float ('Unit cost', digits_compute= dp.get_precision('Sale Price')),
    }

    def onchange_bom_lines_ids (self, cr, uid, ids, bom_lines_ids, product_id, partner_id, qty, pricelist, uom, date_order, context=None):
        context = context or {}
        product_obj = self.pool.get ('product.product')

        res = {}        
        if bom_lines_ids:
            new_unit_price = 0.0
            new_unit_cost = 0.0
            for line in [l[2] for l in bom_lines_ids]:
                new_unit_price += line['unit_price'] * line['quantity']
                new_unit_cost += line['unit_cost'] * line['quantity']

            res ['unit_cost'] = new_unit_cost
            res ['price_unit'] = new_unit_price
            return {'value': res}
        else:
            if product_id:
                #If no bom_lines, the default product price is asumed
                warning = {} 
                product = product_obj.browse(cr, uid, product_id, context=context)
                price = self.pool.get('product.pricelist').price_get(cr, uid, [pricelist],
                                                                     product, qty or 1.0, partner_id, {
                                                                     'uom': uom,
                                                                     'date': date_order,
                                                           })[pricelist]
                if price is False:
                    warning = {
                        'title': 'No valid pricelist line found !',
                        'message':
                            "Couldn't find a pricelist line matching this product and quantity.\n"
                            "You have to change either the product, the quantity or the pricelist."
                        }
                else:
                    res['price_unit'] = price
       
                return {'value': res, 'warning': warning}

        return False


    def product_id_change(self, cr, uid, ids, pricelist, product_id, qty=0,
            uom=False, qty_uos=0, uos=False, name='', partner_id=False,
            lang=False, update_tax=True, date_order=False, packaging=False, fiscal_position=False, flag=False, costlist=None):

        res = super (sale_order_line, self).product_id_change (cr, uid, ids, price_list, product_id, qty, uom, qty_uos,
                                                                uos, name, partner_id, lang, update_tax, date_order, packaging, fiscal_position, flag, costlist)

        if res and res['value'] and costlist:
            product_obj = self.pool.get ('product.product')
            product = product_obj.browse (cr, uid, product_id)
            unit_cost = self.pool.get('product.pricelist').price_get(cr, uid, [costlist],
                    product, qty or 1.0, partner_id, {
                        'uom': uom,
                        'date': date_order,
                        })[costlist]
            if price is False:
                warning = {
                    'title': 'No valid cost pricelist line found !',
                    'message':
                        "Couldn't find a cost pricelist line matching this product and quantity.\n"
                        "You have to change either the product, the quantity or the pricelist."
                    }
            else:
                res['value'].update({'cost_unit': price})

        res['value']['flat_bom_lines_ids'] = []

        return res


    def _create_bom (self, cr, uid, product, bom, count, context=None):

        if product.bom_lines:
            for component in product.bom_lines_ids:
                bom.append ({
                    'product': component.product_id,
                    'label': component.label,
                    'name': component.name,
                    'quantity': component.quantity * count,
                })
        else:
            bom.append({
                'product': product,
                'label': '',
                'name': '',
                'quantity': count,
            })

        return

    def action_compute_flat_bom (self, cr, uid, ids, context=None):
        context = context or {}
        price_list_obj = self.pool.get ('product.pricelist')

        assert ids and len(ids)==1

        sol = self.browse (cr, uid. ids[0], context=context)

        to_delete = []
        for bl in sol.flat_bom_lines_ids:
            to_delete.append((2, bl.id))

        #remove existing boms
        self.write (cr, uid, [sol.id], {'flat_bom_lines_ids': to_delete}, context=context)

        new_bom = []
        self._create_bom (cr, uid, sol.product_id, new_bom, 1.0, context=context)

        consolidated_bom = {}        
        for v in new_bom:
            if v.id in consolidated_bom:
                consolidated_bom[v.id]['quantity'] += v['quantity']
            else:
                consolidated_bom[v.id] = v

        vals = []
        for v in consolidated_bom.values():
            unit_price = price_list_obj.price_get(cr, uid, [sol.order_id.pricelist_id.id],
                                                     v['product'].id,
                                                     v['quantity'] or 1.0, 
                                                     sol.order_id.partner_id.id, 
                                                     {'uom': v['product'].product_uom.id,
                                                      'date': sol.order_id.date_order,
                                                     })[pricelist]
        
            unit_cost = price_list_obj.price_get(cr, uid, [sol.order_id.costlist_id.id],
                                                     v['product'].id,
                                                     v['quantity'] or 1.0, 
                                                     sol.order_id.partner_id.id, 
                                                     {'uom': v['product'].product_uom.id,
                                                      'date': sol.order_id.date_order,
                                                     })[pricelist]

            if not unit_price:
                warning = {
                    'title': 'No valid pricelist line found !',
                    'message':
                        "Couldn't find a pricelist line matching this product and quantity.\n"
                        "You have to change either the product, the quantity or the pricelist."
                    }
            
            if not unit_cost:
                warning = {
                    'title': 'No valid cost pricelist line found !',
                    'message':
                        "Couldn't find a COST pricelist line matching this product and quantity.\n"
                        "You have to change either the product, the quantity or the cost pricelist."
                    }
            

            vals.append([
                (0, 0, {
                    'product_id': v['product'].id, 
                    'quantity': v['quantity'],
                    'unit_cost': unit_cost,
                    'unit_price': unit_price,
                    'margin': unit_price - unit_cost,
                    'margin_percentage': unit_price and (unit_price - unit_cost)/unit_price or 0.0,
            })

        self.write (cr, uid, [sol.id], {'flat_bom_lines_ids': vals}, context=context)

        return False

    def action_change_price_per_category (seld, cr, uid, ids, context=None):
        context = context or {}

        return False

sale_order_line()

